@@ -1,4 +1,3 @@
-/* eslint-env browser */
 const widgets = require('@jupyter-widgets/base');
 const reglScatterplot = require('regl-scatterplot/dist/regl-scatterplot.js');
 const pubSub = require('pub-sub-es');
@@ -6,6 +5,7 @@ const d3Axis = require('d3-axis');
 const d3Scale = require('d3-scale');
 const d3Selection = require('d3-selection');
 const codecs = require('./codecs');
+const createLegend = require('./legend');
 const packageJson = require('../package.json');
 
 const createScatterplot = reglScatterplot.default;
@@ -13,28 +13,37 @@ const createRenderer = reglScatterplot.createRenderer;
 
 const JupyterScatterModel = widgets.DOMWidgetModel.extend(
   {
-    defaults: {
-      ...widgets.DOMWidgetModel.prototype.defaults(),
-      _model_name : 'JupyterScatterModel',
-      _model_module : packageJson.name,
-      _model_module_version : packageJson.version,
-      _view_name : 'JupyterScatterView',
-      _view_module : packageJson.name,
-      _view_module_version : packageJson.version
-    }
+    defaults: Object.assign(
+      {},
+      widgets.DOMWidgetModel.prototype.defaults(),
+      {
+        _model_name : 'JupyterScatterModel',
+        _model_module : packageJson.name,
+        _model_module_version : packageJson.version,
+        _view_name : 'JupyterScatterView',
+        _view_module : packageJson.name,
+        _view_module_version : packageJson.version,
+      }
+    )
   },
   {
-    serializers: {
-      ...widgets.DOMWidgetModel.serializers,
-      points: new codecs.Numpy2D('float32'),
-      selection: new codecs.Numpy1D('uint32'),
-      view_data: new codecs.Numpy1D('uint8'),
-    }
-  },
+    serializers: Object.assign(
+      {},
+      widgets.DOMWidgetModel.serializers,
+      {
+        points: new codecs.Numpy2D('float32'),
+        selection: new codecs.Numpy1D('uint32'),
+        view_data: new codecs.Numpy1D('uint8'),
+      }
+    )
+  }
 );
 
+const AXES_LABEL_SIZE = 16;
 const AXES_PADDING_X = 40;
+const AXES_PADDING_X_WITH_LABEL = AXES_PADDING_X + AXES_LABEL_SIZE;
 const AXES_PADDING_Y = 20;
+const AXES_PADDING_Y_WITH_LABEL = AXES_PADDING_Y + AXES_LABEL_SIZE;
 
 function camelToSnake(string) {
   return string.replace(/[\w]([A-Z])/g, function(m) {
@@ -42,14 +51,6 @@ function camelToSnake(string) {
   }).toLowerCase();
 }
 
-function flipObj(obj) {
-  return Object.entries(obj).reduce((ret, entry) => {
-    const [ key, value ] = entry;
-    ret[ value ] = key;
-    return ret;
-  }, {});
-}
-
 function downloadBlob(blob, name) {
   const link = document.createElement('a');
   link.href = URL.createObjectURL(blob);
@@ -80,8 +81,6 @@ function getScale(scaleType) {
   return d3Scale.scaleLinear();
 }
 
-const MIN_WIDTH = 240;
-
 /**
  * This dictionary maps between the camelCased Python property names and their
  * JavaScript counter parts. In most cases the name is identical but they can be
@@ -98,6 +97,7 @@ const properties = {
   colorSelected: 'pointColorActive',
   colorBy: 'colorBy',
   colorHover: 'pointColorHover',
+  width: 'width',
   height: 'height',
   lassoColor: 'lassoColor',
   lassoInitiator: 'lassoInitiator',
@@ -127,7 +127,14 @@ const properties = {
   viewReset: 'viewReset',
   hovering: 'hovering',
   axes: 'axes',
+  axesColor: 'axesColor',
   axesGrid: 'axesGrid',
+  axesLabels: 'axesLabels',
+  legend: 'legend',
+  legendSize: 'legendSize',
+  legendColor: 'legendColor',
+  legendPosition: 'legendPosition',
+  legendEncoding: 'legendEncoding',
   xScale: 'xScale',
   yScale: 'yScale',
 };
@@ -143,6 +150,7 @@ const reglScatterplotProperty = new Set([
   'pointColorActive',
   'colorBy',
   'pointColorHover',
+  'width',
   'height',
   'lassoColor',
   'lassoInitiator',
@@ -249,6 +257,7 @@ const JupyterScatterView = widgets.DOMWidgetView.extend({
 
       if (self.model.get('axes')) self.createAxes();
       if (self.model.get('axes_grid')) self.createAxesGrid();
+      if (self.model.get('legend')) self.showLegend();
 
       // Listen to events from the JavaScript world
       self.pointoverHandlerBound = self.pointoverHandler.bind(self);
@@ -309,6 +318,25 @@ const JupyterScatterView = widgets.DOMWidgetView.extend({
     this.model.save_changes();
   },
 
+  getOuterDimensions: function getOuterDimensions() {
+    let xPadding = 0;
+    let yPadding = 0;
+
+    if (this.model.get('axes')) {
+      const labels = this.model.get('axes_labels');
+      xPadding = labels ? AXES_PADDING_X_WITH_LABEL : AXES_PADDING_X;
+      yPadding = labels ? AXES_PADDING_Y_WITH_LABEL : AXES_PADDING_Y;
+    }
+
+    const outerWidth = this.model.get('width') === 'auto'
+      ? this.container.getBoundingClientRect().width
+      : this.model.get('width') + xPadding;
+
+    const outerHeight = this.model.get('height') + yPadding;
+
+    return [outerWidth, outerHeight]
+  },
+
   createAxes: function createAxes() {
     this.axesSvg = d3Selection.select(this.container).select('svg').node()
       ? d3Selection.select(this.container).select('svg')
@@ -323,21 +351,25 @@ const JupyterScatterView = widgets.DOMWidgetView.extend({
       .map(function (c) { return Math.round(c * 255); });
     this.axesSvg.style('color', `rgba(${color[0]}, ${color[1]}, ${color[2]}, 1)`);
 
-    const { width, height } = this.container.getBoundingClientRect();
+    const [width, height] = this.getOuterDimensions();
 
     const currentXScaleRegl = this.scatterplot.get('xScale');
     const currentYScaleRegl = this.scatterplot.get('yScale');
 
+    const labels = this.model.get('axes_labels');
+    const xPadding = labels ? AXES_PADDING_X_WITH_LABEL : AXES_PADDING_X;
+    const yPadding = labels ? AXES_PADDING_Y_WITH_LABEL : AXES_PADDING_Y;
+
     // Regl-Scatterplot's gl-space is always linear, hence we have to pass a
     // linear scale to regl-scatterplot.
     // In the future we might integrate this into regl-scatterplot directly
     this.xScaleRegl = d3Scale.scaleLinear()
       .domain(this.model.get('x_domain'))
-      .range([0, width - AXES_PADDING_X]);
+      .range([0, width - xPadding]);
     // This scale is used for the D3 axis
     this.xScaleAxis = getScale(this.model.get('x_scale'))
       .domain(this.model.get('x_domain'))
-      .range([0, width - AXES_PADDING_X]);
+      .range([0, width - xPadding]);
     // This scale converts between the linear, log, or power normalized data
     // scale and the axis
     this.xScaleRegl2Axis = getScale(this.model.get('x_scale'))
@@ -346,10 +378,10 @@ const JupyterScatterView = widgets.DOMWidgetView.extend({
 
     this.yScaleRegl = d3Scale.scaleLinear()
       .domain(this.model.get('y_domain'))
-      .range([height - AXES_PADDING_Y, 0]);
+      .range([height - yPadding, 0]);
     this.yScaleAxis = getScale(this.model.get('y_scale'))
       .domain(this.model.get('y_domain'))
-      .range([height - AXES_PADDING_Y, 0]);
+      .range([height - yPadding, 0]);
     this.yScaleRegl2Axis = getScale(this.model.get('y_scale'))
       .domain(this.model.get('y_domain'))
       .range(this.model.get('y_domain'));
@@ -374,7 +406,7 @@ const JupyterScatterView = widgets.DOMWidgetView.extend({
       : this.axesSvg.append('g').attr('class', 'x-axis');
 
     this.xAxisContainer
-      .attr('transform', `translate(0, ${height - AXES_PADDING_Y})`)
+      .attr('transform', `translate(0, ${height - yPadding})`)
       .call(this.xAxis);
 
     this.yAxisContainer = this.axesSvg.select('.y-axis').node()
@@ -382,21 +414,52 @@ const JupyterScatterView = widgets.DOMWidgetView.extend({
       : this.axesSvg.append('g').attr('class', 'y-axis');
 
     this.yAxisContainer
-      .attr('transform', `translate(${width - AXES_PADDING_X}, 0)`)
+      .attr('transform', `translate(${width - xPadding}, 0)`)
       .call(this.yAxis);
 
     this.axesSvg.selectAll('.domain').attr('opacity', 0);
 
+    if (labels) {
+      this.xAxisLabel = this.axesSvg.select('.x-axis-label').node()
+        ? this.axesSvg.select('.x-axis-label')
+        : this.axesSvg.append('text').attr('class', 'x-axis-label');
+
+      this.xAxisLabel
+        .text(labels[0])
+        .attr('text-anchor', 'middle')
+        .attr('font-size', '12px')
+        .attr('font-weight', 'bold')
+        .attr('x', (width - xPadding) / 2)
+        .attr('y', height);
+
+      this.yAxisLabel = this.axesSvg.select('.y-axis-label').node()
+        ? this.axesSvg.select('.y-axis-label')
+        : this.axesSvg.append('text').attr('class', 'y-axis-label');
+
+      this.yAxisLabel
+        .text(labels[1])
+        .attr('text-anchor', 'middle')
+        .attr('dominant-baseline', 'hanging')
+        .attr('x', (height - yPadding) / 2)
+        .attr('y', -width)
+        .attr('font-size', '12px')
+        .attr('font-weight', 'bold')
+        .attr('transform', `rotate(90)`);
+    }
+
+    this.updateContainerDimensions();
+
     this.scatterplot.set({
       xScale: this.xScaleRegl,
       yScale: this.yScaleRegl,
-      height: this.model.get('height') - AXES_PADDING_Y,
     });
 
-    this.canvasWrapper.style.right = `${AXES_PADDING_X}px`;
-    this.canvasWrapper.style.bottom = `${AXES_PADDING_Y}px`;
+    this.canvasWrapper.style.right = `${xPadding}px`;
+    this.canvasWrapper.style.bottom = `${yPadding}px`;
 
     if (this.model.get('axes_grid')) this.createAxesGrid();
+
+    this.updateLegendWrapperPosition();
   },
 
   removeAxes: function removeAxes() {
@@ -406,16 +469,20 @@ const JupyterScatterView = widgets.DOMWidgetView.extend({
     this.yAxis = undefined;
     this.xAxisContainer = undefined;
     this.yAxisContainer = undefined;
+    this.xAxisContainer = undefined;
+    this.xAxisLabel = undefined;
+    this.yAxisLabel = undefined;
 
     this.canvasWrapper.style.top = '0';
     this.canvasWrapper.style.left = '0';
     this.canvasWrapper.style.right = '0';
     this.canvasWrapper.style.bottom = '0';
 
+    this.updateContainerDimensions();
+
     this.scatterplot.set({
       xScale: undefined,
       yScale: undefined,
-      height: this.model.get('height'),
     });
   },
 
@@ -452,27 +519,140 @@ const JupyterScatterView = widgets.DOMWidgetView.extend({
     }
   },
 
+  showLegend: function showLegend() {
+    this.hideLegend();
+
+    this.legendWrapper = document.createElement('div');
+    this.legendWrapper.className = 'legend-wrapper';
+    this.legendWrapper.style.position = 'absolute';
+    this.legendWrapper.style.pointerEvents = 'none';
+    this.updateLegendWrapperPosition();
+
+    this.legend = createLegend(
+      this.model.get('legend_encoding'),
+      this.model.get('legend_color'),
+      this.model.get('background_color'),
+      this.model.get('legend_size')
+    );
+    this.updateLegendPosition();
+
+    this.legendWrapper.appendChild(this.legend);
+    this.container.appendChild(this.legendWrapper);
+  },
+
+  hideLegend: function hideLegend() {
+    if (!this.legendWrapper) return;
+    this.container.removeChild(this.legendWrapper);
+    this.legendWrapper = undefined;
+    this.legend = undefined;
+  },
+
+  updateLegendWrapperPosition: function updateLegendWrapperPosition() {
+    if (!this.legendWrapper) return;
+
+    const labels = this.model.get('axes_labels');
+    const xPadding = labels ? AXES_PADDING_X_WITH_LABEL : AXES_PADDING_X;
+    const yPadding = labels ? AXES_PADDING_Y_WITH_LABEL : AXES_PADDING_Y;
+
+    this.legendWrapper.style.top = 0;
+    this.legendWrapper.style.bottom = yPadding + 'px';
+    this.legendWrapper.style.left = 0;
+    this.legendWrapper.style.right = xPadding + 'px';
+  },
+
+  updateLegendPosition: function updateLegendPosition() {
+    if (!this.legend) return;
+
+    this.legend.style.position = 'absolute';
+    this.legend.style.top = null;
+    this.legend.style.bottom = null;
+    this.legend.style.left = null;
+    this.legend.style.right = null;
+    this.legend.style.transform = null;
+
+    const position = this.model.get('legend_position');
+    let translateX = 0;
+    let translateY = 0;
+
+    if (position.indexOf('top') >= 0) {
+      this.legend.style.top = 0;
+    } else if (position.indexOf('bottom') >= 0) {
+      this.legend.style.bottom = 0;
+    } else {
+      this.legend.style.top = '50%';
+      translateY = '-50%';
+    }
+
+    if (position.indexOf('left') >= 0) {
+      this.legend.style.left = 0;
+    } else if (position.indexOf('right') >= 0) {
+      this.legend.style.right = 0;
+    } else {
+      this.legend.style.left = '50%';
+      translateX = '-50%';
+    }
+
+    if (translateX || translateY) {
+      this.legend.style.transform = `translate(${translateX}, ${translateY})`;
+    }
+  },
+
+  updateContainerDimensions: function updateContainerDimensions() {
+    const width = this.model.get('width');
+    const height = this.model.get('height');
+
+    let xPadding = 0;
+    let yPadding = 0;
+
+    if (this.model.get('axes')) {
+      const labels = this.model.get('axes_labels');
+      xPadding = labels ? AXES_PADDING_X_WITH_LABEL : AXES_PADDING_X;
+      yPadding = labels ? AXES_PADDING_Y_WITH_LABEL : AXES_PADDING_Y;
+    }
+
+    this.container.style.width = width === 'auto'
+      ? '100%'
+      : (width + xPadding) + 'px';
+    this.container.style.height = (height + yPadding) + 'px';
+
+    window.requestAnimationFrame(() => { this.resizeHandler(); });
+  },
+
   resizeHandler: function resizeHandler() {
     if (!this.model.get('axes')) return;
 
-    const { width, height } = this.container.getBoundingClientRect();
+    const [width, height] = this.getOuterDimensions();
 
-    this.xScaleAxis.range([0, width - AXES_PADDING_X]);
-    this.yScaleAxis.range([height - AXES_PADDING_Y, 0]);
+    const labels = this.model.get('axes_labels');
+    const xPadding = labels ? AXES_PADDING_X_WITH_LABEL : AXES_PADDING_X;
+    const yPadding = labels ? AXES_PADDING_Y_WITH_LABEL : AXES_PADDING_Y;
+
+    this.xScaleAxis.range([0, width - xPadding]);
+    this.yScaleAxis.range([height - yPadding, 0]);
     this.xAxis.scale(this.xScaleAxis);
     this.yAxis.scale(this.yScaleAxis);
 
     this.xAxisContainer
-      .attr('transform', `translate(0, ${height - AXES_PADDING_Y})`)
+      .attr('transform', `translate(0, ${height - yPadding})`)
       .call(this.xAxis);
     this.yAxisContainer
-      .attr('transform', `translate(${width - AXES_PADDING_X}, 0)`)
+      .attr('transform', `translate(${width - xPadding}, 0)`)
       .call(this.yAxis);
 
+    this.updateLegendWrapperPosition();
+
+    this.withPropertyChangeHandler('width', this.model.get('width') || 'auto');
+    this.withPropertyChangeHandler('height', this.model.get('height'));
+
     // Render grid
     if (this.model.get('axes_grid')) {
-      this.xAxis.tickSizeInner(-(height - AXES_PADDING_Y));
-      this.yAxis.tickSizeInner(-(width - AXES_PADDING_X));
+      this.xAxis.tickSizeInner(-(height - yPadding));
+      this.yAxis.tickSizeInner(-(width - xPadding));
+    }
+
+    if (labels) {
+      this.xAxisLabel.attr('x', (width - xPadding) / 2).attr('y', height);
+      this.yAxisLabel.attr('x', (height - yPadding) / 2).attr('y', -width);
     }
   },
 
@@ -568,11 +748,11 @@ const JupyterScatterView = widgets.DOMWidgetView.extend({
     }
   },
 
-  xScaleHandler: function xScaleHandler(newXScale) {
+  xScaleHandler: function xScaleHandler() {
     this.createAxes();
   },
 
-  yScaleHandler: function yScaleHandler(newXScale) {
+  yScaleHandler: function yScaleHandler() {
     this.createAxes();
   },
 
@@ -615,13 +795,12 @@ const JupyterScatterView = widgets.DOMWidgetView.extend({
     }
   },
 
-  heightHandler: function heightHandler(newValue) {
-    this.container.style.height = newValue + 'px';
-    this.withPropertyChangeHandler(
-      'height',
-      this.model.get('axes') ? newValue - AXES_PADDING_Y : newValue
-    );
-    this.resizeHandler();
+  widthHandler: function widthHandler() {
+    this.updateContainerDimensions();
+  },
+
+  heightHandler: function heightHandler() {
+    this.updateContainerDimensions();
   },
 
   backgroundColorHandler: function backgroundColorHandler(newValue) {
@@ -755,11 +934,44 @@ const JupyterScatterView = widgets.DOMWidgetView.extend({
     else this.removeAxes();
   },
 
+  axesColorHandler: function axesColorHandler() {
+    this.createAxes();
+  },
+
   axesGridHandler: function axesGridHandler(newValue) {
     if (newValue) this.createAxesGrid();
     else this.removeAxesGrid();
   },
 
+  axesLabelsHandler: function axesLabelsHandler(newValue) {
+    if (!newValue) this.removeAxes();
+    this.createAxes();
+  },
+
+  legendHandler: function legendHandler(newValue) {
+    if (newValue) this.showLegend();
+    else this.hideLegend();
+  },
+
+  legendColorHandler: function legendColorHandler() {
+    this.hideLegend();
+    this.showLegend();
+  },
+
+  legendSizeHandler: function legendSizeHandler() {
+    this.hideLegend();
+    this.showLegend();
+  },
+
+  legendPositionHandler: function legendPositionHandler() {
+    this.updateLegendPosition();
+  },
+
+  legendEncodingHandler: function legendEncodingHandler() {
+    if (!this.model.get('legend')) return;
+    this.showLegend();
+  },
+
   otherOptionsHandler: function otherOptionsHandler(newOptions) {
     this.scatterplot.set(newOptions);
   },
